/***************************************************************************
 * qlocx_temp.c 				                                                
 *											
 *  
 *  
 *                                                                                      
 * Functional level: 1									
 *											
 * Prepared: Dirk Handzic							
 *																					
 * 											
 * Rev	Date	Author	Ext int. (Y/N) 	Reason/description
 * 0.4  210114  DH      N               disable NTC power after use
 * 0.3  201218  DH      Y               EN NTC power added
 * 0.2  201214  DH      N               read correct adc channel
 * 0.1  201029  DH	Y        	Initial version
 *
 *  
 * Temperature input handler 										
 ***************************************************************************/


#include <net/socket.h>
#include <net/lwm2m.h>
#include <nrf9160.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <drivers/adc.h>
#include <drivers/sensor.h>
#include <zephyr.h>
#include <logging/log.h>
#include "qlocx_temp.h"
#include "qlocx_adc.h"
#include "qlocx_digio.h"

LOG_MODULE_REGISTER(app_qlocx_temp, CONFIG_APP_LOG_LEVEL);



// ADC values for 10k NTC with B=3380 and 10k fixed resistor -35 deg C to +120 deg C
#define NTC_TABLE_SIZE  32
#define MIN_TEMP        -350
#define MAX_TEMP        1200
#define ONE_DEG         10

const uint16_t temp[NTC_TABLE_SIZE] =  {3518, 3452, 3372, 3278, 3168,       //-35, -30,- 25, -20, -15
                                        3044, 2904, 2751, 2586, 2413,       // -10, -5, 0, 5, 10
                                        2235, 2055, 1877, 1703, 1573,       // 15, 20, 25, 30, 35 
                                        1381, 1236, 1102, 981, 871,         // 40, 45, 50, 55, 60
                                        772, 685, 607, 538, 476,            // 65, 70, 75, 80, 85
                                        423, 375, 333, 297, 264,            // 90, 95, 100, 105, 110
                                        236, 211};                          // 115, 120


// Calculate temp from ADC value using linear interpolation from TEMP table

int16_t qlocx_CalcTemp(uint16_t value)
{
int16_t result;
uint8_t index;
uint8_t done;
uint16_t delta;


     // Return min or max value
    if(value > temp[0]){
        return(MIN_TEMP - ONE_DEG);
    }
    if(value < temp[NTC_TABLE_SIZE - 1]){
        return(MAX_TEMP + ONE_DEG);
    }
    // if value within limits calculate the correct temperature
    done = 0;
    index = NTC_TABLE_SIZE - 1;
    result = MAX_TEMP;
    do{
        if(value <= temp[index]){
            done = 1;
        }
        else{
            index--;
            result -= 50;
        }
     }while((index < (NTC_TABLE_SIZE - 1)) && (done == 0));


    delta = (temp[index] - temp[index + 1]);
    result += ((temp[index] - value)*50)/delta;
    return(result);
}

/* use 25.5C if no sensor available */
static struct float32_value temp_float = { 25, 500000 };
#if ADD_TIMESTAMPS
static int32_t timestamp;
#endif




int read_temperature( struct float32_value *float_val)
{

int temp;
struct sensor_value temp_val;


        temp = qlocx_CalcTemp(qlocx_get_adc(0));

        temp_val.val1 = temp / 10;
        temp_val.val2 = fabs((temp - (temp_val.val1 * 10)) * 100000);
#ifdef CONSOLE_PRINT
	printk("NTC: read %d.%d C\n", temp_val.val1, temp_val.val2);
#endif
	float_val->val1 = temp_val.val1;
	float_val->val2 = temp_val.val2;

	return 0;
}




static void *temp_read_cb(uint16_t obj_inst_id, uint16_t res_id, uint16_t res_inst_id,
			  size_t *data_len)
{
	int32_t ts;
        int err;


	/* Only object instance 0 is currently used */
	if (obj_inst_id != 0) {
		*data_len = 0;
		return NULL;
	}


	qlocx_digio_NTCPwrOn(true);
        k_sleep(K_MSEC(10));
        err = adc_sample();
	read_temperature(&temp_float);
        qlocx_digio_NTCPwrOn(false);
	lwm2m_engine_set_float32("3303/0/5700", &temp_float);
	*data_len = sizeof(temp_float);
        qlocx_digio_NTCPwrOn(false);
        #if ADD_TIMESTAMPS
	/* get current time from device */
	lwm2m_engine_get_s32("3/0/13", &ts);
        /* set timestamp */
	lwm2m_engine_set_s32("3303/0/5518", ts);
        #endif
	return &temp_float;
}



int qlocx_init_temp(void)
{


	
        // Temperature object
	lwm2m_engine_create_obj_inst("3303/0");
	lwm2m_engine_register_read_callback("3303/0/5700", temp_read_cb);
        #if ADD_TIMESTAMPS
	lwm2m_engine_set_res_data("3303/0/5518",
				  &timestamp, sizeof(timestamp), 0);

        #endif
        qlocx_offset_adc();

	return 0;
}


float32_value_t qlocx_test_temp(void)
{
int err;

    err = adc_sample();
    read_temperature(&temp_float);
    return(temp_float);

}